/*
 * ar4c-dcdc - main.c
 *
 * Copyright (C) 2014  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <leds.h>
#include <comm.h>
#include <adc.h>
#include <tio.h>
#include <jiffies.h>
#include <watchdog.h>
#include <storage.h>
#include <pins.h>
#include <a2f.h>
#include <limits.h>

/* 1000 quants per Volt */
#define V_Q	1000

#define MV_TO_Q(x)	((unsigned int)(((unsigned long)(x) * V_Q) / 1000))

/*
 * used for hysteresis
 */
#define V_TINY		MV_TO_Q(20)
#define V_SMALL		MV_TO_Q(50)
#define V_MED		MV_TO_Q(100)

/*
 * Minimal DC input voltage for normal operation. The DC converter
 * needs at least 18V. We add 1V for a diode.
 */
#define VIN_MIN		MV_TO_Q(19000)

/*
 * Maximal DC input voltage for normal operation. If the voltage
 * is too high something is really wrong.
 */
#define VIN_MAX		MV_TO_Q(35000)

/*
 * Battery Controller parameters
 *
 * Bellow VBAT_PRESENT_MIN we assume that there are no battery.
 * Bellow VBAT_CHARGER_MIN we disable charger to prevent overheat and
 * immediately report Battery Failure status.
 * We have three low battery levels - VBAT_MIN, VBAT_CRITICAL and VBAT_LOW.
 * Below VBAT_MIN we always shutdown the DC/DC converter. Between
 * VBAT_MIN and VBAT_CRITICAL we inform host about critical battery
 * level and wait for shutdown confirmation. The VBAT_LOW is only used
 * just to inform about low battery level.
 *
 * The VBAT_MAX is mostly used for checking for failures.
 */
#define VBAT_PRESENT_MIN	MV_TO_Q(2000)
#define VBAT_CHARGER_MIN	MV_TO_Q(10000)	/* 5V per battery */
#define VBAT_MIN		MV_TO_Q(21800)	/* 10.9V */
#define VBAT_CRITICAL		MV_TO_Q(22800)	/* 11.4V */
#define VBAT_LOW		MV_TO_Q(23600)	/* 11.8V */
#define VBAT_CHARGER_LOW	MV_TO_Q(25200)	/* 12.6V, 2.1V per cell */
#define VBAT_CHARGER_SP		MV_TO_Q(27400)	/* 13.7V, 13.6V-13.8V */
#define VBAT_MAX		MV_TO_Q(31000)	/* 15.5V */

#define CHARGING_DELAY		(5 * HZ)

#define CHARGING_TIME_MAX	(100 * 3600UL * HZ)
#define CHARGING_TIME_FLOAT	(8 * 3600UL * HZ)
#define CHARGING_TIME_LOW_FLOAT	(12 * 3600UL * HZ)

#define RECHARGE_TIME_MAX	(28 * 24 * 3600UL * HZ)
#define SELFDISCHARGE_MIN	(10 * 24 * 3600UL * HZ)
#define BROKEN_RECHARGE_TIME	(12 * 3600UL * HZ)

/*
 * The BD244 transistor has 1 W thermal dissipation limit. If the
 * battery voltage is too low we can easily exceed that limit.
 *
 * We have hardware 0.25 A current controller and we limit the power
 * on transistor to 0.75 W - we set THERMAL_DISSIPATION to 3 V * 0.25 A
 * = 0.75 W.
 */
#if HZ <= 100
#define THERMAL_SHIFT		(3)
#else
#define THERMAL_SHIFT		(6)
#endif
#define THERMAL_DISSIPATION	(MV_TO_Q(3000) >> THERMAL_SHIFT)
#define THERMAL_Q_LIMIT		(THERMAL_DISSIPATION * 2 * HZ)

static unsigned int dc_update_status(unsigned int vin, unsigned int vbat,
		unsigned int flags)
{
	/*
	 * Check DC input voltage. It should higher tha VIN_MIN.
	 */
	if (vin > VIN_MIN + V_MED)
		flags |= DC_DC0_PRESENT;
	else if (vin < VIN_MIN - V_MED)
		flags &= ~DC_DC0_PRESENT;

	/*
	 * Check if the DC input voltage is too high.
	 */
	if (vin > VIN_MAX + V_MED)
		flags |= DC_DC0_HIGH;
	else if (vin < VIN_MAX - V_MED)
		flags &= ~DC_DC0_HIGH;

	return flags;
}

enum {
	BATTERY_INIT,
	BATTERY_DISCHARGING,
	BATTERY_WANT_CHARGE,
	BATTERY_CHARGING,
	BATTERY_CHARGED,
	BATTERY_BROKEN,
	BATTERY_SHUTDOWN,
};
static unsigned char battery_state;
static unsigned long batstate_ts;

static void set_state(unsigned char state)
{
	battery_state = state;
	batstate_ts = get_jiffies();
}

static u16 config_dirty;

static void config_touch(unsigned char reg)
{
	config_dirty |= 1 << (reg - CONFIG_REGS_BASE);
}

/*
 * Voltage based charging termination
 *
 * If the voltage will be not increased in the next VBAT_UP_TIME_MAX
 * jiffies the charging will stop. Additionally if we exceed VBAT_CHARGER_SP
 * we always clear this timer.
 */
#define VBAT_UP_TIME_MAX	(5 * 60 * HZ)
#define VBAT_UP_C_MAX		(6)
static unsigned int vbat_max;
static unsigned int vbat_max_c;
static unsigned long vbat_ts;
static unsigned long vbat_f;

/*
 * The DC/DC works at 61 Hz,
 * for 3.6 Ah battery the increase of voltage during chaging was
 * ab. 4 mV/minute. So the ~1/32 s bandwidth should be adequate.
 *
 * So we use x1024 oversampling (about 5 extra bits)
 */
#define VBAT_FILTER_SHIFT	10
static void vbat_filter(unsigned int vbat)
{
	long t;

	t = vbat - (long)(vbat_f >> VBAT_FILTER_SHIFT);
	vbat_f += t;
	comm_reg_write(COMM_VBAT_F, vbat_f >> 10);
}

static void update_selfdischarge_rate(u16 rate)
{
	u16 old = comm_reg_read(COMM_SELFDISCHARGE);

	if (rate != old) {
		comm_reg_write(COMM_SELFDISCHARGE, rate);
		config_touch(COMM_SELFDISCHARGE);
	}
}

/*
 * estimate_ic - estimate current charging electrical current
 *
 * Ic = throttle * Imax
 * Ib = throttle * Imax
 * where:
 *	Imax = min(I_MAX, (vin - vbat) / 2.5)
 */
static unsigned char estimate_ic(unsigned int vin, unsigned int vbat,
		unsigned char enabled)
{
	unsigned char ic = 0;
	unsigned int vd;

	if ((vin > vbat) && enabled) {
		/*
		 * The nominal current Imax is 250 mA, but it's limitted to
		 * 0.4 * (vin - vbat), approximated with 0.4% error
		 */
		vd = vin - vbat;
		ic = 250;
		if (vd < 625)
			ic = (vd * 102) >> 8;
	}
	return ic;
}

static unsigned char battery_charging_prev;
static unsigned long charging_end_ts;
static unsigned char charging_flags;
#define CHARGING_SP_REACHED		0x01
#define CHARGING_MINDIFF_REACHED	0x02

static unsigned int battery_charging(unsigned int vin, unsigned int vbat,
		unsigned int flags, unsigned long t)
{
	static struct a2f_state ibat_filter;
	unsigned int vbat_sp;
	unsigned char charger_enable = 0;
	unsigned int throttle;
	unsigned char ibat;
	unsigned char state;
	unsigned int vb;
	unsigned int q;
	int x;

	flags &= ~DC_BAT_CHARGING;
	if (vin + V_MED < vbat) {
		state = BATTERY_DISCHARGING;
		goto stop_charging;
	}

	if (vbat < VBAT_CHARGER_MIN) {
		state = BATTERY_INIT;
		goto stop_charging;
	}

	if (time_after(t, charging_end_ts)) {
		if (!(charging_flags & CHARGING_SP_REACHED))
			flags |= DC_BAT_CHARGE_ERR;
		state = BATTERY_CHARGED;
		goto stop_charging;
	}

	vb = comm_reg_read(COMM_VBAT_F);
	if ((vb > vbat_max) || (vbat >= VBAT_CHARGER_SP)) {
		vbat_max = vb;
		vbat_ts = t;
	}

	if (time_after(t, vbat_ts + VBAT_UP_TIME_MAX)) {
		if (vbat_max_c >= VBAT_UP_C_MAX) {
			state = BATTERY_BROKEN;
			goto stop_charging;
		}
		vbat_max_c++;
		vbat_max = vb;
		vbat_ts = t;
	}

	vbat_sp = VBAT_CHARGER_SP;

	charger_enable = 1;
	q = comm_reg_read(COMM_Q);
	if (vbat >= vbat_sp) {
		if (!(charging_flags & CHARGING_SP_REACHED)) {
			charging_flags |= CHARGING_SP_REACHED;
			charging_end_ts = t + CHARGING_TIME_FLOAT;
		}
		charger_enable = 0;
		goto out;
	}

	if (q >= THERMAL_Q_LIMIT) {
		charger_enable = 0;
		goto out;
	}

	/*
	 * add a short rest pulse. It should help battery a little.
	 */
	if (!((unsigned char) jiffies & 0x3f)) {
		charger_enable = 0;
		goto out;
	}

	/*
	 * if the battery voltage is too high, disable charging for a while.
	 * It could be just one sample, but we want to be sure that the
	 * battery really exists. We abuse the thermal limit to have higher
	 * delay.
	 *
	 * We also need to touch the battery voltage-based overcharing
	 * protection because we expect that the battery voltage will fall
	 * because of the throttling, not a cell failure.
	 */
	if (vin < vbat + V_MED) {
		static unsigned char counter;

		vbat_max = vb;
		vbat_ts = t;
		counter++;
		if (!counter) {
			if (!charging_flags) {
				charging_end_ts = t + CHARGING_TIME_LOW_FLOAT;
				charging_flags |= CHARGING_MINDIFF_REACHED;
			}

			q = THERMAL_Q_LIMIT + THERMAL_DISSIPATION * (HZ / 2);
			charger_enable = 0;
			goto out;
		}
	}

	if (charger_enable) {
		int vd = (int)(vin >> THERMAL_SHIFT)
				- (vbat >> THERMAL_SHIFT);

		if (vd > 16384)
			vd = 16384;
		else if (vd < 0)
			vd = 0;
		q += vd;
	}
out:
	q -= THERMAL_DISSIPATION;
	if (q > UINT_MAX - THERMAL_DISSIPATION)
		q = 0;
	comm_reg_write(COMM_Q, q);

	throttle = comm_reg_read(COMM_THROTTLE);
	x = (charger_enable << 14) - throttle;
	if (x >= 0)
		x += 128;
	else
		x -= 128;
	throttle += x >> 8;

	a2f_step(&ibat_filter, 8, estimate_ic(vin, vbat, battery_charging_prev));
	ibat = a2f_get(&ibat_filter, 8);

	flags |= DC_BAT_CHARGING;

	comm_reg_write(COMM_THROTTLE, throttle);
	comm_reg_write(COMM_IBAT, ibat);
	battery_charger(charger_enable);
	battery_charging_prev = charger_enable;
	return flags;

stop_charging:
	comm_reg_write(COMM_THROTTLE, 0);
	comm_reg_write(COMM_IBAT, 0);
	battery_charger(0);
	battery_charging_prev = 0;
	a2f_reset(&ibat_filter);
	set_state(state);
	return flags;
}

/*
 * Battery controller
 *
 * This is the core part of battery controller. This code sets following
 * flags:
 *
 * FLAG				UPDATE
 * -----------------------------------------------
 * DC_BAT_HIGH			ALWAYS
 * DC_BAT_PRESENT		ALWAYS
 * DC_BAT_CHARGE_ERR		ALWAYS
 * DC_BAT_DISCHARGING		BATTERY_DISCHARGING
 * DC_BAT_CRITICAL		BATTERY_DISCHARGING
 * DC_BAT_LOW			BATERRY_DISCHARGING
 * DC_BAT_CHARGING		BATTERY_CHARGING
 * DC_SHUTDOWN			BATTERY_SHUTDOWN
 * DC_BAT_GOOD			BATTERY_CHARGED
 */
unsigned int battery_control(unsigned int flags)
{
	unsigned int vin = comm_reg_read(COMM_VIN);
	unsigned int vbat = comm_reg_read(COMM_VBAT);
	unsigned long t = get_jiffies();

	/*
	 * Check if the DC input voltage is too high.
	 */
	if (vbat > VBAT_MAX + V_MED)
		flags |= DC_BAT_HIGH;
	else if (vbat + V_MED < VBAT_MAX)
		flags &= ~DC_BAT_HIGH;

	if (vbat > VBAT_PRESENT_MIN + V_MED) {
		flags |= DC_BAT_PRESENT;

		/*
		 * For proper charging we need at least VBAT_CHARGER_SP
		 * DC input voltage. Otherwise the battery will be probably
		 * undercharged and may fail prematurely.
		 */
		if (vin > VBAT_CHARGER_SP + 3 * V_MED)
			flags &= ~DC_DC0_LOW;
		else if (vin < VBAT_CHARGER_SP + V_MED)
			flags |= DC_DC0_LOW;
	} else if (vbat < VBAT_PRESENT_MIN - V_MED) {
		flags &= ~(DC_BAT_PRESENT | DC_BAT_FAIL | DC_DC0_LOW);
		update_selfdischarge_rate(0);
	}

	switch (battery_state) {
	/*
	 * BATTERY_INIT - we don't know anything about battery. If something
	 * unexpected occurs we go here.
	 */
	case BATTERY_INIT:
		if (vin + V_MED < vbat)
			set_state(BATTERY_DISCHARGING);
		else if ((vin >= vbat + V_MED) && (vbat >= VBAT_CHARGER_MIN))
			set_state(BATTERY_WANT_CHARGE);
		break;

	/*
	 * BATTERY_DISCHARGING - normal battery discharging, the DC voltage
	 * is lower than the battery voltage (usually the DC is not present)
	 *
	 * We just monitor the battery state and the DC voltage.
	 */
	case BATTERY_DISCHARGING:
		if (vin > vbat + V_MED) {
			flags &= ~DC_BAT_DISCHARGING_CLEAR;
			set_state(BATTERY_INIT);
		} else if (vbat < VBAT_MIN) {
			flags &= ~DC_BAT_DISCHARGING_CLEAR;
			set_state(BATTERY_SHUTDOWN);
		} else {
			flags |= DC_BAT_DISCHARGING;

			if (vbat < VBAT_CRITICAL - V_SMALL)
				flags |= DC_BAT_CRITICAL;
			else if (vbat > VBAT_CRITICAL + V_SMALL)
				flags &= ~DC_BAT_CRITICAL;

			if (vbat < VBAT_LOW - V_SMALL)
				flags |= DC_BAT_LOW;
			else if (vbat > VBAT_LOW + V_SMALL)
				flags &= ~DC_BAT_LOW;
		}
		break;

	/*
	 * BATTERY_SHUTDOWN - datalogger is powered down
	 *
	 * We power up the datalogger if the DC is back and it's also
	 * above battery voltage.
	 */
	case BATTERY_SHUTDOWN:
		flags &= ~DC_SHUTDOWN;
		if ((vin >= VIN_MIN) && (vin > vbat + V_MED))
			set_state(BATTERY_INIT);
		else
			flags |= DC_SHUTDOWN;
		break;

	/*
	 * BATTERY_WANT_CHARGE - we want to charge battery
	 */
	case BATTERY_WANT_CHARGE:
		if (vin + V_MED < vbat) {
			set_state(BATTERY_DISCHARGING);
		} else if (vin < vbat + V_MED) {
			set_state(BATTERY_INIT);
		} else if (time_after(t, batstate_ts + CHARGING_DELAY)) {
			set_state(BATTERY_CHARGING);
			vbat_max_c = 0;
			vbat_max = vbat;
			vbat_ts = t;
			charging_end_ts = t + CHARGING_TIME_MAX;
			charging_flags = 0;
			flags |= DC_BAT_CHARGING;
		}
		break;

	/*
	 * BATTERY_CHARGING - battery charging is enabled
	 */
	case BATTERY_CHARGING:
		flags = battery_charging(vin, vbat, flags, t);
		break;

	/*
	 * BATTERY_CHARGED - batterry is fully charged
	 *
	 * We have expected voltage the DC is present, we just check for
	 * DC and self-discharge.
	 */
	case BATTERY_CHARGED:
		flags &= ~DC_BAT_GOOD;
		if (vin + V_MED < vbat) {
			set_state(BATTERY_DISCHARGING);
		} else if (vbat < VBAT_MIN) {
			set_state(BATTERY_INIT);
		} else if (vbat < VBAT_CHARGER_LOW) {
			u16 rate = (t - batstate_ts) / (60 * HZ);

			update_selfdischarge_rate(rate + 1);
			if (time_after(t, batstate_ts + SELFDISCHARGE_MIN))
				flags &= ~DC_BAT_FAIL;
			else
				flags |= DC_BAT_FAIL;

			set_state(BATTERY_WANT_CHARGE);
		} else if (time_after(t, batstate_ts + RECHARGE_TIME_MAX)) {
			u16 rate = RECHARGE_TIME_MAX / (60 * HZ);

			update_selfdischarge_rate(rate + 1);
			flags &= ~DC_BAT_FAIL;
			set_state(BATTERY_WANT_CHARGE);
		} else {
			flags |= DC_BAT_GOOD;
		}
		break;

	/*
	 * BATTERY_BROKEN - battery is broken
	 */
	case BATTERY_BROKEN:
		flags &= ~DC_BAT_GOOD;
		flags |= DC_BAT_FAIL;
		if (vin + V_MED < vbat)
			set_state(BATTERY_DISCHARGING);
		else if (vbat < VBAT_CHARGER_MIN)
			set_state(BATTERY_INIT);
		else if (time_after(t, batstate_ts + BROKEN_RECHARGE_TIME))
			set_state(BATTERY_WANT_CHARGE);
		break;
	}

	if (!(flags & DC_BAT_CHARGING))
		battery_charger(0);

	return flags;
}

static unsigned int dcdc_exec_commands(unsigned int flags)
{
	unsigned int cmd = comm_reg_read(COMM_COMMAND);

	if (cmd) {
		if (!(flags & DC_BAT_CRITICAL))
			cmd &= ~COMM_CMD_POWEROFF_ALL;

		if ((cmd & COMM_CMD_POWEROFF_ALL) == COMM_CMD_POWEROFF_ALL) {
			flags &= ~DC_BAT_SHUTDOWN_CLEAR;
			set_state(BATTERY_SHUTDOWN);
		}

		comm_reg_clear(COMM_COMMAND, cmd);
	}
	return flags;
}

static void config_load_all(void)
{
	unsigned char i;

	for (i = 0; i < CONFIG_REGS_COUNT; i++)
		comm_reg_write(CONFIG_REGS_BASE + i, config_read(i));
}

static unsigned int initialize_flags(void)
{
	unsigned int flags = 0;
	u16 x;

	x = comm_reg_read(COMM_SELFDISCHARGE);
	if (x && (x <= SELFDISCHARGE_MIN / (60 * HZ)))
		flags |= DC_BAT_FAIL;
	return flags;
}

static void config_updated(void)
{
	unsigned char i;
	u16 mask = 1;
	u16 x;

	if (!config_dirty)
		return;

	for (i = 0; i < CONFIG_REGS_COUNT; i++, mask <<= 1) {
		if (config_dirty & mask) {
			config_dirty &= ~mask;
			x = comm_reg_read(CONFIG_REGS_BASE + i);
			config_write(x, i);
			return;
		}
	}
	config_dirty = 0;
}

int main(void)
{
	unsigned long t;
	unsigned int flags;
	unsigned int vin;
	unsigned int vbat;

	watchdog_init();
	pins_init();
	adc_init();
	comm_init();
	jiffies_init();
	set_state(BATTERY_INIT);

	wdt_reset();
	config_load_all();
	comm_reg_write(COMM_VERSION, 7);
	tio_init();

	flags = initialize_flags();

	t = get_jiffies();
	for (;;) {
		wdt_reset();
		tio_read();

		vin = comm_reg_read(COMM_VIN);
		vbat = comm_reg_read(COMM_VBAT);
		flags = dc_update_status(vin, vbat, flags);
		vbat_filter(vbat);
		flags = battery_control(flags);
		flags = dcdc_exec_commands(flags);

		comm_reg_write(COMM_FLAGS, flags);

		dcdc_shutdown(battery_state == BATTERY_SHUTDOWN);

		switch (battery_state) {
		case BATTERY_CHARGING:
			if (t & 0x40)
				led_set(POWER_LED);
			else
				led_clear(POWER_LED);
			break;
		case BATTERY_CHARGED:
		case BATTERY_SHUTDOWN:
			led_clear(POWER_LED);
			break;
		case BATTERY_INIT:
		case BATTERY_DISCHARGING:
		case BATTERY_WANT_CHARGE:
			led_set(POWER_LED);
			break;
		}

		t++;
		config_updated();
		sleep_to(t);
	}
}
